/******************************************************************************
*                                                                             *
*                          Network Job Control (NJC)                          *
*                                                                             *
*                             Q U E U E D . C                                 *
*                                 queued                                      *
*                               Version 2.0                                   *
*                                                                             *
*                   Copyright (c) Joerg-R. Hill, April 1992                   *
*                                                                             *
*-----------------------------------------------------------------------------*
* This is the server for Network Job Control. It has to run on one host in    *
* the network.                                                                *
* Possible command line options: -debug, -time <n>                            *
*-----------------------------------------------------------------------------*
* Revision history: 1.0 first version using rsh for networking and signals    *
*                       for interprocess communication                        *
*                   2.0 use of RPC implemented and now working only in        *
*                       connection with queuec                                *
******************************************************************************/

#include<errno.h>
#include<pwd.h>
#include<signal.h>
#include<stdio.h>
#include<string.h>
#include<time.h>
#include<unistd.h>
#ifdef AIX
#include<sys/select.h>
#endif
#include<sys/stat.h>
#include<sys/types.h>
#include<sys/wait.h>
#include<rpc/rpc.h>

#include "queue.h"

/* Prototypes */
void main(int, char **);
void init(void);
void next_job(void);
void update_queue(int, FILE *);
int  submit_job(char *);
int  find_compi(char *);
void terminate(int);
void ende(int);
void stop(void);
void svc_run_new(void);
int  *queueok(void);

/* Prototypes from util.c */
extern FILE *open_file(char *, char *, char *);
extern FILE *reopen_file(char *, char*, char *, FILE *);
extern int  setmode(char *, char *, int);
extern int  move_file(char *, char *, char *);
extern int  rm(char *, char *);
extern int  update_file(FILE *, char *);
extern int  exist(char *, char *);

/* Prototypes from checkpids.c */
extern int  *checkpids(int *);

/* Prototypes from read_pids.c */
extern void read_config(void);
extern void read_pids(void);

/* Prototypes from error_msg.c */
extern void error_msg(int, char *);

/* Prototypes from xdr.c */
extern bool_t xdr_intarr(XDR *, int *);
extern bool_t xdr_job(XDR *, JOB *);

/* Structures and global variables used */

struct {
  char name[MAXLENHOSTNAME];
  int  maxjobs;
  int  maxallowed;
  int  nice;
  int  load;
  int  numpids;
  int  pids[MAXPROCESS];
} host[MAXHOSTS];

struct timeval sleeptime = {600, 0};

int numhosts, debug=FALSE, first=TRUE, stopped=FALSE, locked=FALSE;
int print_date=TRUE;
char *prgname, hostname[MAXLENHOSTNAME], queue_server[MAXLENHOSTNAME];

void main(int argc, char *argv[])       /* Here the main programme starts */
{ 
  int chpid;
  register int i;

  prgname=strrchr(argv[0],'/');         /* Find programme name */
  if (prgname == NULL) prgname=argv[0];
  else prgname++;

  if (getuid())         /* Check whether root has started the programme */
  {
    fprintf(stderr,"%s: ERROR: Only root can execute this programme.\n",
			 prgname);
    exit(-1);
  }

  for (i=1; i<argc; i++)                        /* Look for options */
  {
    if (!strcmp(argv[i],"-debug")) debug=TRUE;  /* debug option     */
    if (!strcmp(argv[i],"-time"))
    {
	i++;
	sleeptime.tv_sec=atoi(argv[i]);           /* time option      */
	if (sleeptime.tv_sec <= 0)
	{
	  fprintf(stderr, "%s: ERROR: Wrong argument for -time option.\n",
				 prgname);
        exit(-1);
      }
    }
  }

  /* Fork a new process so that the programme started from the terminal
     can be completed */
  if ((chpid=fork()) < 0)
  {
    error_msg(2, "");
    exit(-1);
  }
  else if (chpid == 0)
  {
    init();

    signal(SIGTERM, terminate);            /* catch terminate signal */ 

    svc_run_new();                         /* listen on network port */
  }
}

void init(void)
{                                    /* this function reads the configuration */
  FILE *cfgfile;                     /* from file, it is also called when     */
  register int i=0;                  /* configuration changes are required    */
  char line[MAXLENLINE];

  if (first)                         /* we are for the first time in init()   */
  {                                  /* look for LOCK file                    */
    strcpy(line, NJCPATH);
    strcat(line, "/LOCK");
    if ((cfgfile=fopen(line,"r")) != NULL)  /* if it exists -> exit */
    {
	fprintf(stderr,"%s: ERROR: %s is already running.\n", prgname, prgname);
	fclose(cfgfile);
	exit(-1);
    }                                                     /* create LOCK file */
    if ((cfgfile=open_file(NJCPATH, "LOCK", "w")) == NULL) exit(-1);
    fclose(cfgfile);
    if (!setmode(NJCPATH, "LOCK", S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) exit(-1);

    gethostname(hostname, MAXLENHOSTNAME);
    strcpy(queue_server, hostname);

    /* Register RPC routines */
    if ((registerrpc(NJCSERVER, VERS, NEXTJOB, (void *)next_job, xdr_void, xdr_void)) == -1)
    {
      fprintf(stderr, "%s: ERROR: Could not register service.\n", prgname);
	exit(-1);
    }
    if ((registerrpc(NJCSERVER, VERS, INIT, (void *)init, xdr_void, xdr_void)) == -1)
    {
      fprintf(stderr, "%s: ERROR: Could not register service.\n", prgname);
	exit(-1);
    }
    if ((registerrpc(NJCSERVER, VERS, STOP, (void *)stop, xdr_void, xdr_void)) == -1)
    {
      fprintf(stderr, "%s: ERROR: Could not register service.\n", prgname);
	exit(-1);
    }
    if ((registerrpc(NJCSERVER, VERS, QUEUESERVER, (void *)queueok, xdr_void, xdr_int)) == -1)
    {
      fprintf(stderr, "%s: ERROR: Could not register service.\n", prgname);
	exit(-1);
    }
    first=FALSE;
  }

  if (reopen_file(NJCPATH, "stderr.log", "a", stderr) == NULL) exit(-1);

  read_config();

  if (debug)                           /* print configuration if -debug was */
  {                                    /* specified                         */
    printf("\nHost\tJobs\tNice\n");
    for (i=0; i<=numhosts; i++)
	printf("%s\t%d\t%d\n", host[i].name, host[i].maxjobs, host[i].nice);
  }
}

void next_job(void)
{                               /* this function is called when a new job */
  FILE *queue, *j;              /* shall be started                       */
  register int rmlin, success;
  char line[MAXLENLINE];

  locked=TRUE;                  /* lock call to next_job in main programme  */

  read_pids();                  /* read queue.running file into memory */
                                /* and check correctness               */
  if (debug) printf("%s: Trying to start next job.\n", prgname);

  while (exist(NJCPATH, "lock"))      /* Open queue.waiting, make sure that */
    sleep(3);                         /* only one process does it           */
  if ((j=open_file(NJCPATH, "lock", "w")) != NULL) fclose(j);

  if ((queue=open_file(NJCPATH, "queue.waiting", "r+")) == NULL) return;
								    /* look into queue */

  rmlin=0;
  success=FALSE;
  while((fgets(line, MAXLENLINE, queue)) != NULL)
  {
    rmlin++;                                         /* there is a job */
    if (submit_job(line))                            /* try to submit  */
    {
      if (debug) printf("%s: Job started\n", prgname);
      update_queue(rmlin, queue);             /* remove job from queue */
	success=TRUE;
      break;
    }
  }
  if (!success) fclose(queue);
  rm(NJCPATH, "lock");

  if (debug)
    if (!rmlin) printf("%s: No more jobs.\n", prgname);
}

void update_queue(int rmlin, FILE *queue) /* this function removes a job just */
{                                         /* submitted from the queue         */
  FILE *newqueue;
  char tmp[L_tmpnam];
  register int i;
  char line[MAXLENLINE];

  rewind(queue);
  if ((newqueue=open_file("", tmp, "w")) == NULL) return;

  clearerr(newqueue);
  i=0;
  while((fgets(line, MAXLENLINE, queue)) != NULL)
  {
    i++;
    if (i != rmlin)
	fprintf(newqueue, "%s", line);
  }
  i=ferror(newqueue);
  fclose(newqueue);
  if (i)
  {
    error_msg(12, tmp);
    return;
  } 
  if (!update_file(queue, tmp)) return;
}

int submit_job(char *line)
{                                      /* this function submits a job */
  JOB job;
  struct passwd *entry;
  char compi[MAXLENLINE], *word;
  long at;
  int istat;
  register int i;

  word=strtok(line, ";");
  if (word == NULL) return(FALSE);

  job.user=word;               /* break line from queue.waiting file into */
  word=strtok(NULL, ";");      /* fields                                  */
  strcpy(compi, word);
  word=strtok(NULL, ";");
  at=atol(word);

  if (at > (time(NULL)+30) || at < 0)
  {
    if (debug) printf("%s: Job has still to wait: Start time %ld, now it is %ld\n",
			     prgname, at, time(NULL));
    return(FALSE);
  }                            /* job shall still wait */

  word=strtok(NULL, ";");
  word=strtok(NULL, ";");
  job.path=word;
  word=strtok(NULL, ";");
  job.command=word;
  if ((word=strrchr(job.command, '\n')) != NULL) *word='\0';

  if ((i=find_compi(compi)) == -1) return(FALSE);     /* no host free, return */

  if (debug) printf("%s: Submitting to %s.\n", prgname, host[i].name);

  entry=getpwnam(job.user);             /* Get user password from /etc/passwd */
  job.password=entry->pw_passwd;      /* Fill password and shell in structure */
  job.shell=strrchr(entry->pw_shell, '/')+1;                         /* "job" */
  job.nice=host[i].nice;       /* Fill host specific field in structure "job" */

  if ((istat=callrpc(host[i].name, NJCCLIENT, VERS, STARTJOB, (xdrproc_t)xdr_job,
                     (char *)(&job), (xdrproc_t)xdr_void, (char *)0)) != 0)
  {                                 /* Call queuec on remote host via RPC */
    error_msg(5, host[i].name);
      clnt_perrno(istat);                /* An error occured, return with */
    fprintf(stderr, "\n");               /* this bad message              */
    return(FALSE);
  }
  return(TRUE);                          /* O.K. job submitted, return    */
}

int find_compi(char *compi)             /* this function looks for a free */
{                                       /* host                           */
  int load[MAXHOSTS], minload;
  char *word;
  register int i, j;

  for (i=0; i<=numhosts; i++)
    load[i]=100;

  if (*compi != '*')                  /* only some hosts are allowed, set */
  {                                   /* "load" for these hosts to the    */
    word=strtok(compi,",");           /* value from "host[].load"         */
    if (word != NULL)
    {
	for (i=0; i<=numhosts; i++)
	{
	  if (!strcmp(word, host[i].name))
	  {
	    load[i]=host[i].load;
	    break;
        }
      }
	while ((word=strtok(NULL, ",")) != NULL)
	{
	  for (i=0; i<=numhosts; i++)
  	  {
	    if (!strcmp(word, host[i].name))
	    {
	      load[i]=host[i].load;
	      break;
          }
        }
      }
    }
  }
  else                                /* all hosts are allowed, set "load"  */
    for (i=0; i<=numhosts; i++) load[i]=host[i].load;

  for (i=0; i<=numhosts; i++)
  {                                   /* check how many jobs are running on */
    if (load[i] == 0)                 /* "host[i].name"                     */
      return(i);                      /* no jobs -> the best result we      */
  }                                   /* can expect, use this host          */

  minload=100;                        /* check which host has the lowest */
  j=0;                                /* "load"                          */
  for (i=0; i<=numhosts; i++)
  {
    if (load[i] < minload)
    {
	minload=load[i];
	j=i;
    }
  }
  if (minload == 100)                 /* no host found which will accept a */
    return(-1);                       /* job now                           */
  else
    return(j);                        /* use host with minimum "load" */
}

void terminate(int dummy)                 /* this function is called when */
{                                         /* SIGTERM was catched          */
  ende(0);
}

void ende(int ret)
{                                         /* this function is called when    */
  rm(NJCPATH, "LOCK");                    /* queued has to exit (SIGTERM     */ 
  exit(ret);                              /* was catched or an error occured */
}                                         /* remove LOCK file then exit      */

void stop(void)
{
  FILE *msgfile;                          /* this function is called when  */
                                          /* the queue shall be started or */
  if (stopped)                            /* stopped                       */
  {
    stopped=FALSE;
    if (debug) printf("%s: Queue has been started.\n", prgname);
    error_msg(8, "");
    if (!rm(NJCPATH, "queue.stopped")) return;
  }
  else
  {
    stopped=TRUE;
    if (debug) printf("%s: Queue has been stopped.\n", prgname);
    error_msg(9, "");
    if ((msgfile=open_file(NJCPATH, "queue.stopped", "w")) == NULL) return;
    fclose(msgfile);
    setmode(NJCPATH, "queue.stopped",
       	S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
  }
}

void svc_run_new(void)           /* This function is an own version of */
{                                /* svc_run(), it returns after        */
  fd_set readfds;                /* "sleeptime.tv_sec" seconds         */
  int dtbsz=getdtablesize();

  FD_ZERO(&readfds);
  for (;;)
  {
    readfds.fds_bits[0]=svc_fds;
    switch (select(dtbsz, &readfds, NULL, NULL, &sleeptime))
    {
	case -1:
  	  if (errno == EINTR) continue;
	  error_msg(11, "");
	  return;
      case 0:
	  if (!locked && !stopped) next_job(); /* don't try to start next job if it  */
	  locked=FALSE;                        /* was just tried because the new job */
        break;                               /* isn't yet registered               */
	default:
	  svc_getreqset(&readfds);
    }
  }
}

int *queueok(void) /* This function is called via RPC from */
{                  /* the clients to find out the server   */
  static int pid;

  pid=-getpid();
  return(&pid);
}
