/******************************************************************************
*                                                                             *
*                          Network Job Control (NJC)                          *
*                                                                             *
*                            S T A R T J O B . C                              *
*                                                                             *
*                   Copyright (c) Joerg-R. Hill, April 1992                   *
*                                                                             *
******************************************************************************/

#include<pwd.h>
#include<signal.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<unistd.h>
#include<sys/param.h>
#include<sys/stat.h>
#ifdef AIX
#include<sys/select.h>
#endif
#include<sys/times.h>
#include<sys/types.h>
#include<sys/errno.h>
#include<sys/wait.h>
#include<rpc/rpc.h>

#include "queue.h"

/* Prototypes */
void startjob(JOB *);
void execute_job(JOB *);
int  forkjob(JOB *);
int  sendmail(char *, char *, char *);
int  cputime(void);

/* Prototypes from util.c */
extern FILE *open_file(char *, char*, char *);
extern FILE *reopen_file(char *, char*, char *, FILE *);
extern int  setmode(char *, char *, int);
extern int  move_file(char *, char *, char *);
extern int  rm(char *, char *);
extern int  exist(char *, char *);
extern int  filelength(char *, char *);
extern int  update_file(FILE *, char *);

/* Prototypes from error_msg.c */
extern void error_msg(int, char *);

/* Global variables */
extern char *prgname;
extern char queue_server[MAXLENHOSTNAME];
extern int  debug;
extern char hostname[MAXLENHOSTNAME];

static char mailhost[MAXLENHOSTNAME];

void startjob(JOB *job)        /* This function is called via RPC from queued */
{
  FILE *file;
  struct passwd *entry;
  static char home[MAXLENLINE];
  static char user[L_cuserid+6];
  static char logname[L_cuserid+9];
  char line[MAXLENLINE], timestr[26], *h;
  pid_t chpid;
  time_t now;

  signal(SIGCHLD, SIG_IGN);     /* Don't wait for childrens termination, this */
                                /* would produce zombies only                 */
  entry=getpwnam(job->user);    /* Get entry from passwd corresponding to     */
                                /* user name and check password               */
  gethostname(hostname, MAXLENHOSTNAME);

  if (strcmp(job->password, entry->pw_passwd))
  {
    error_msg(13, job->user);
    if ((file=open_file("", line, "w")) != NULL)
    {
      now=time(NULL);
      strcpy(timestr, ctime(&now));
      timestr[24]='\0';
	fprintf(file, "Subject: Message from Network Job Control\nTo: %s@%s\n\n",
		  job->user, mailhost);
      fprintf(file, "\n*** Message from Network Job Control ***\n");
      fprintf(file, "***     %s     ***\n\n", timestr);
      fprintf(file, "Job      : %s\n", job->command);
      fprintf(file, "Directory: %s\n", job->path);
      fprintf(file, "Host     : %s\n\n", hostname);
      fprintf(file, "The above mentioned job cannot be executed. Permission is denied.\n");
      fclose(file);
      sendmail(job->user, hostname, line);
      return;
    }
  }

  if ((chpid=fork()) < 0)       /* Fork new process where we can change UID   */
    error_msg(2, "");
  else if (chpid == 0)
  {
    if (setgid(entry->pw_gid) || setreuid(entry->pw_uid,entry->pw_uid))
    {                                                      /* set GID and UID */
      error_msg(4, job->user);
      exit(-1);
    }

    signal(SIGCHLD, SIG_DFL);        /* Now we are waiting again for children */

    if (strlen(entry->pw_dir)+6 <= MAXLENLINE)  /* Generate environment, HOME */
    {
      strcpy(home, "HOME=");
      strcat(home, entry->pw_dir);
      putenv(home);
    }
    else
    {
      fprintf(stderr, "%s: Path name for HOME directory exceeds %d characters.\n", prgname, MAXLENLINE);
      exit(-1);
    }
    putenv("TERM=no_term");                     /* and TERM */
    strcpy(user, "USER=");                      /* and USER */
    strcat(user, job->user);
    putenv(user);
    strcpy(logname, "LOGNAME=");                /* and LOGNAME */
    strcat(logname, job->user);
    putenv(logname);
    job->shell=strrchr(entry->pw_shell, '/')+1;

    strcpy(mailhost, hostname);         /* Looking for 'mailhost=' in $HOME/.queuerc */
    if (exist(entry->pw_dir, ".queuerc"))
    {
      if (debug) fprintf(stderr, "%s: %s/.queuerc file found.\n", prgname,
                         entry->pw_dir);
      if ((file=open_file(entry->pw_dir, ".queuerc", "r")) != NULL)
      {
        while (fgets(line, MAXLENLINE, file) != NULL)
        {
          if (*line == '#') continue;
          if (strstr(line, "mailhost"))
          {
            h=strchr(line, '=')+1;        /* We have found it, use this host for mail  */
              if (strstr(h, "nomail"))    /* If entry was mailhost=nomail send no mail */
                mailhost[0]='\0';
              else
              strcpy(mailhost, h);
            if ((h=strrchr(mailhost, '\n')) != NULL) *h='\0';
          }
        }
        fclose(file);
      }
    }
    if (debug)
    {
      if (mailhost[0] != '\0')
        fprintf(stderr, "%s: Mail will go to host %s.\n", prgname, mailhost);
      else
        fprintf(stderr, "%s: No mail will be sent.\n", prgname); 
    }

    execute_job(job);

    exit(0);
  }  /* end of if (chpid=fork() ... */
  return;
}

void execute_job(JOB *job)       /* This function executes the desired job */
{
  static FILE *j, *k;
  char line[MAXLENLINE], saveline[MAXLENLINE], *h;
  char tmp[L_tmpnam], timestr[26];
  int istat, cpu;
  time_t now, wall;
  pid_t pid;
  register int i;

  if (chdir(job->path))                     /* go into directory of the job */
  {
    error_msg(10, job->path);
    return;
  }

  if (debug) fprintf(stderr, "%s: Path changed to %s.\n", prgname, job->path);

  now=time(NULL);                           /* What's the time ? */
  pid=getpid();                             /* What's our PID ?  */

  while (exist(NJCPATH, "lock"))    /* Update queue.running, make sure that */
    sleep(3);                       /* only one process does it             */
  if ((j=open_file(NJCPATH, "lock", "w")) != NULL) fclose(j);

  if ((j=open_file(NJCPATH, "queue.running", "a")) == NULL) return;
  fprintf(j, "%s;%s;%d;%ld;%s;%s\n", job->user, hostname, pid, time(NULL),
          job->path, job->command);         /* Update queue.running */
  fclose(j);
  rm(NJCPATH, "lock");

  if (debug) fprintf(stderr, "%s: Queue.running updated.\n", prgname);

  if ((j=open_file(NJCPATH, "time.log", "a")) != NULL)  /* Write message into */
  {                                                     /* time.log file      */
    fprintf(j, "%s\t%s\t%d\tstarted\t%s", job->user, hostname, pid, ctime(&now));
    fclose(j);
    if (debug) fprintf(stderr, "%s: Time.log updated.\n", prgname);
  }

  sprintf(saveline, "I am process %d on %s.", pid, hostname); /* Create file to */
  if ((j=open_file(".", saveline, "w")) != NULL) fclose(j);   /* inform user    */
                                                    /* where his job is running */

  istat=forkjob(job);                  /* Now do the real job */
  cpu=cputime();

  rm(".", saveline);                   /* Remove file informing the user */

  wall=now;
  now=time(NULL);
  strcpy(timestr, ctime(&now));
  timestr[24]='\0';
  wall=now-wall;                                /* Determine wall clock time */
  if ((j=open_file(NJCPATH, "time.log", "a")) != NULL) /* Write message into */
  {                                                    /* time.log           */
    fprintf(j,"%s\t%s\t%d\tended  \t%s\tCPU: %d s\tTotal: %ld s\n", job->user, hostname, pid, timestr, cpu, wall);
    fclose(j);
    if (debug) fprintf(stderr, "%s: Time.log updated.\n", prgname);
  }

  while (exist(NJCPATH, "lock"))    /* Update queue.running, make sure that */
    sleep(3);                       /* only one process does it             */
  if ((j=open_file(NJCPATH, "lock", "w")) != NULL) fclose(j);

  if ((j=open_file(NJCPATH, "queue.running", "r+")) == NULL) return;
  if ((k=open_file("", tmp, "w")) == NULL) return;
  clearerr(k);
  while ((fgets(line, MAXLENLINE, j)) != NULL)
  {
    strcpy(saveline, line);
    strtok(line, ";");
    h=strtok(NULL, ";");
    i=atoi(strtok(NULL, ";"));
    if (strcmp(h, hostname) || i != pid)
      fprintf(k, "%s", saveline);
  }
  i=ferror(k); 
  fclose(k);
  if (i)
  {
    error_msg(12, tmp);
    return;
  }
  if (!update_file(j, tmp))
  {
    error_msg(12, "queue.running");
    return;
  }

  rm(NJCPATH, "lock");

  if (debug) fprintf(stderr, "%s: Queue.running updated.\n", prgname);

  if (mailhost[0] != '\0')
  {
    if ((j=open_file("", tmp, "w")) != NULL)    /* Send mail to user */
    {
	fprintf(j, "Subject: Message from Network Job Control\nTo: %s@%s\n\n",
		  job->user, mailhost);
      fprintf(j, "\n*** Message from Network Job Control ***\n");
      fprintf(j, "***     %s     ***\n\n", timestr);
      fprintf(j, "Job      : %s\n", job->command);
      fprintf(j, "Directory: %s\n", job->path);
      fprintf(j, "Host     : %s\n", hostname);
      fprintf(j, "CPU  Time: %d s\n", cpu);
      fprintf(j, "Wall Time: %ld s\n\n", wall);
      fprintf(j, "The above mentioned job ");
      if (istat>1000 && istat<2000)
        fprintf(j, "was stopped by signal %d.\n", istat-1000);
      else if (istat==2000)
        fprintf(j, "completed successfully.\n              Congratulations !\n");
      else if (istat>2000 && istat<3000)
      {
        fprintf(j, "completed with an error. Its exit status is %d", istat-2000);
        switch(istat-2000)
        {
          case ENOENT: fprintf(j, " (no such file or directory).\n");
                       break;
          case EACCES: fprintf(j, " (permission denied).\n");
                       break;
          case ENOSPC: fprintf(j, " (no space left on device).\n");
                       break;
          default:     fprintf(j, ".\n");
                       break;
        }
      }
      else
        fprintf(j, "was terminated by signal %d.\n", istat-3000);
      fclose(j);

      if (sendmail(job->user, mailhost, tmp)) exit(0);
    }
  }

  callrpc(queue_server, NJCSERVER, VERS, NEXTJOB, (xdrproc_t)xdr_void, (char *)0,
          (xdrproc_t)xdr_void, (char *)0);
                     /* Send message to queued that this host may get a new job */

  if (filelength(".", "stdout.log") == 0)   /* Remove stdout.log if it is empty */
    rm(".", "stdout.log");
  if (!debug)
  {
    if (filelength(".", "stderr.log") == 0) /* Remove stderr.log if it is empty */
      rm(".", "stderr.log");
  }
}

int forkjob(JOB *job)     /* This function forks a new process which then exec's */
{                         /* a login shell which executes the job                */
  char command[MAXLENLINE];
  pid_t chpid;
  int status;
                                                      /* Build command line */
  sprintf(command, "nice -%d %s", job->nice, job->command);

  if ((chpid=fork()) < 0)                               /* Fork new process */
  { 
    error_msg(2, "");
    exit(-1);
  }
  else if (chpid == 0)
  {
    if (!debug) reopen_file(job->path, "stderr.log", "w", stderr); /* Redirect */
    reopen_file(job->path, "stdout.log", "w", stdout);     /* input and output */
    reopen_file("/dev", "null", "r", stdin);
                                                          /* and execute shell */
    if (debug) fprintf(stderr, "%s: Executing %s %s -c %s\n", prgname,
                       job->shell, job->shell, command);
    if (execlp(job->shell, job->shell, "-c", command, (char *)0))
    {
      error_msg(6, job->shell);
      exit(-1);
    }
  }
  else
  {                         /* The parent waits for termination of the shell */
    if (debug) fprintf(stderr, "%s: Waiting for process %d.\n", prgname, chpid);
    while((wait(&status)) != chpid);
    if (WIFSTOPPED(status) != 0)  return(1000+WSTOPSIG(status));
    if (WIFEXITED(status) != 0)   return(2000+WEXITSTATUS(status));
    if (WIFSIGNALED(status) != 0) return(3000+WTERMSIG(status));
  }
  return(0);                /* We never arrive here */
}

int sendmail(char *user, char *host, char *file)
{
  int status;
  pid_t chpid;
  char line[MAXLENLINE];

  sprintf(line, "%s@%s", user, host);
  if ((chpid=fork()) < 0)
    error_msg(2, "");
  else if (chpid == 0)
  {
    freopen(file, "r", stdin);       /* Executing mail */
    if (execlp("mail", "mail", line, (char *)0)) error_msg(6, "mail");
    return(-1);
  }
  else
    while((wait(&status)) != chpid);
  unlink(file);                      /* Remove mail file */
  return(0);
}

int cputime(void)            /* This function returns the cpu time used */
{
  struct tms buffer;

  if (times(&buffer) != -1)
    return((buffer.tms_utime + buffer.tms_stime + buffer.tms_cutime + buffer.tms_cstime)/HZ);
  else
    return(-1);
}
