/******************************************************************************
*                                                                             *
*                          Network Job Control (NJC)                          *
*                                                                             *
*                              S U B M I T . C                                *
*                                  submit                                     *
*                                Version 3.0                                  *
*                                                                             *
*                   Copyright (c) Joerg-R. Hill, April 1992                   *
*                                                                             *
*-----------------------------------------------------------------------------*
* This is the user interface for Network Job Control. It has to be available  *
* on every host in the network which is managed by NJC.                       *
* Possible command line option: see subroutine helpme()                       *
*-----------------------------------------------------------------------------*
* Revision history: 1.0    simple shell script based on at command            *
*                   1.1    network capabilities added (-c option)             *
*                   2.0    completely rewritten in C language and now working *
*                          only in connection with queued                     *
*                   2.1    use of curses routines implemented                 *
*                   3.0    use of RPC implemented and now working only in     *
*                          connection with queuec and queued                  *
*******************************************************************************/

#include<ctype.h>
#include<pwd.h>
#include<signal.h>
#include<stdlib.h>
#include<stdio.h>
#include<string.h>
#include<time.h>
#ifndef LINUX
#include<curses.h>
#include<term.h>
#else
#include<ncurses/ncurses.h>
#include<ncurses/term.h>
#endif
#include<unistd.h>
#ifdef AIX
#include<sys/access.h>
#include<sys/select.h>
#endif
#include<sys/stat.h>
#include<sys/types.h>
#include<rpc/rpc.h>

#include "queue.h"

/* Prototypes */
int  main(int, char **);
void init(void);
void helpme(void);
void list_jobs(void);
void show_stat(void);
void rm_job(void);
void hold_job(char *);
void set_queue(int);
void submit_job(void);
time_t conv_time(void);
void check_name(char *);
char *timetostr(time_t);
void sethost(char *, int);
void write_cfg(void);
void allowed(void);
void kill_job(char *, int);
void stop_queue(void);
void force_job(void);
void on_off(int);
void check_time(char *, int dat[5]);
void interactive(void);
void clean_up(int);
void ende(int);
void make_hostlist(int);
int  host_in_list(char *);
void terminate(int);
int menu(char **menue, int, char *, int, int);

/* Prototypes from read_pids.c */
extern void read_config(void);
extern void read_pids(void);

/* Prototypes from util.c */
extern FILE *open_file(char *, char *, char *);
extern int  move_file(char *, char *, char *);
extern int  setmode(char *, char *, int);
extern int  update_file(FILE *, char *);
extern int  rm(char *, char *);
extern int  exist(char *, char *);

/* Prototypes from error_msg.c */
extern void error_msg(int, char *);

/* Global variables used */
struct {
  char name[MAXLENHOSTNAME];
  int  maxjobs;
  int  maxallowed;
  int  nice;
  int  load;
  int  numpids;
  int  pids[MAXPROCESS];
} host[MAXHOSTS];

char *prgname, user[L_cuserid], hostname[MAXLENHOSTNAME];
char compi[MAXLENLINE], job[MAXLENLINE], attime[MAXLENLINE];
char queue_server[MAXLENHOSTNAME]="", *word;
char hostlist[MAXHOSTS][MAXLENHOSTNAME];
uid_t uid;
int jobs=(-1), nicev=(-1), maxjobs=(-1), jobno=0, numhosts;
int list=FALSE, all=FALSE, status=FALSE, help=FALSE, rmj=FALSE;
int stop=FALSE, force=FALSE, moni=FALSE, hold=FALSE, stopped=FALSE;
int off=FALSE, on=FALSE, print_date=FALSE;

int main(int argc, char **argv)   /* the main programme */
{
  register int i;

  prgname=strrchr(argv[0],'/');   /* get name of this programme as it was  */
  if (prgname == NULL) prgname=argv[0];  /* typed at the command line      */
  else prgname++;

  *job='\0';                      /* initialize some variables */
  *attime='\0';
  strcpy(compi,"*");

  signal(SIGINT,  clean_up);
  signal(SIGQUIT, clean_up);
  signal(SIGTERM, clean_up);

  if (argc == 1) interactive();   /* and scan list of command line options */
  for (i=1; i<=argc-1; i++)
  {
    if (*argv[i] == '-')          /* it is an option */
    {
      if (!strcmp(argv[i],"-l"))
        list=TRUE;
      else if (!strcmp(argv[i],"-la") || !strcmp(argv[i],"-al"))
      {
        list=TRUE;
        all=TRUE;
      }
      else if (!strcmp(argv[i],"-t"))
        strcpy(attime,argv[++i]);
      else if (!strcmp(argv[i],"-c"))
        strcpy(compi,argv[++i]);
      else if (!strcmp(argv[i],"-s"))
        status=TRUE;
      else if (!strcmp(argv[i],"-h"))
        help=TRUE;
      else if (!strcmp(argv[i],"-r"))
      {
        rmj=TRUE;
        jobno=atoi(argv[++i]);
      }
      else if (!strcmp(argv[i],"-jobs"))
        jobs=atoi(argv[++i]);
      else if (!strcmp(argv[i],"-maxjobs"))
        maxjobs=atoi(argv[++i]);
      else if (!strcmp(argv[i],"-nice"))
        nicev=atoi(argv[++i]);
      else if (!strcmp(argv[i],"-stop") || !strcmp(argv[i],"-start"))
        stop=TRUE;
      else if (!strcmp(argv[i],"-off"))
        off=TRUE;
      else if (!strcmp(argv[i],"-on"))
        on=TRUE;
      else if (!strcmp(argv[i],"-f"))
        force=TRUE;
      else if (!strcmp(argv[i],"-m"))
        moni=TRUE;
      else if (!strcmp(argv[i],"-hold") || !strcmp(argv[i],"-rel"))
      {
        hold=TRUE;
        jobno=(++i);
      }
      else
      {
        fprintf(stderr,"%s: ERROR: Unknown option %s.\n",prgname,argv[i]);
        exit(-1);
      }
    }
    else
    {                               /* no option --> append it to job */
      strcat(job, argv[i]);
      strcat(job, " ");
    }
  }

  init();
  if (help) helpme();                         /* do the requested job */
  else if (list) list_jobs();
  else if (status) show_stat();
  else if (rmj) rm_job();
  else if (hold) hold_job(argv[jobno]);
  else if (jobs != -1) set_queue(0);
  else if (nicev != -1) set_queue(1);
  else if (maxjobs != -1) set_queue(2);
  else if (stop) stop_queue();
  else if (force) force_job();
  else if (on) on_off(TRUE);
  else if (off) on_off(FALSE);
  else submit_job();
  return(0);
}

void init(void)                  /* this function reads the queue.cfg file */
{                                /* and fills the host-structure           */
  struct passwd *entry;
  int res;
  register int i=0;

  gethostname(hostname, MAXLENHOSTNAME);    /* get hostname of this machine */

  uid=getuid();
  entry=getpwuid(uid);
  strcpy(user, entry->pw_name);             /* get user name of the user who */
                                            /* currently runs submit         */
  read_config();

  for (i=0; i<=numhosts; i++)
  {
    if (callrpc(host[i].name, NJCSERVER, VERS, QUEUESERVER, (xdrproc_t)xdr_void,
                (char *)0, (xdrproc_t)xdr_int, (char *)(&res)) == RPC_SUCCESS)
    {
      strcpy(queue_server, host[i].name);
      break;
    }
  }
  if (*queue_server == '\0')
  {
    fprintf(stderr, "%s: ERROR: No queue server found.\n", prgname);
    exit(-1);
  }

  read_pids();

  if (exist(NJCPATH, "queue.stopped")) stopped=TRUE;
}

void helpme(void)               /* this function does what its name promises */
{
  static char *helptext[] = {
  "  Usage: submit <job>                          submits a job now",
  "         submit <job> -t <time>                submits a job at time",
  "         submit -c <hostname>[,<hostname>...]  submit works only on hosts",
  "                                               hostname ... (default: all)",
  "         submit -c ^<hostname>[,<hostname>...] submit works on all hosts",
  "                                               expect hostname ...",
  "         submit -l                             shows jobs of user in",
  "                                               the queue",
  "         submit -la                            shows all jobs in the",
  "                                               queue",
  "         submit -r <n>                         removes job n from the queue",
  "         submit -hold <n>[[,<n> ...][-<n>]]    sets job(s) n ... to hold",
  "         submit -rel <n>[[,<n> ...][-<n>]]     release job(s) n ...",
  "         submit -f                             force queued to try to start",
  "                                               next job",
  "         submit -s                             shows queue parameters",
  "         submit -jobs <n>                      sets queue to run n jobs in",
  "                                               parallel",
  "         submit -maxjobs <n>                   sets queue to allow n jobs in",
  "                                               parallel",
  "         submit -nice <n>                      sets queue to submit jobs with",
  "                                               a nice of n",
  "         submit -h                             shows help for the submit command",
  "         submit -start | -stop                 starts or stops the queue",
  "         submit -off | -on                     switch host off/on for queueing",
  " ",
  "  <time> may be: <day> or <day>,<hh>:<mm> or <month>/<date>/<year>,<hh>:<mm> or",
  "                 <month>/<date>/,<hh>:<mm> or <hh>:<mm> or '+<integer> <time_unit>'",
  "  <day> may be:  Mon, Tue, Wed, Thu, Fri, Sat, Sun, tomorrow, hold",
  "  <time_unit> may be: minutes, hours, days, weeks, months, years",
  "*"
  };
  char *term;
  int curses=TRUE;
  register int j=0, i=0;

  term=getenv("TERM");
  if ((newterm(term, stdout, stdin)) == (SCREEN *)0) curses=FALSE;

  if (curses)
  {
    signal(SIGINT,  terminate);        /* handle quit signals in own routine */
    signal(SIGQUIT, terminate);
    signal(SIGTERM, terminate);
    cbreak();
    noecho();
    if (COLS < 80)
    {
      endwin();
      fprintf(stderr, "%s: ERROR: Use a window with more than 79 columns.\n", prgname);
      exit(-1);
    }

    move(i++, 0);
    addstr(prgname);
  }
  else
    printf("%s\n", prgname);

  while (*helptext[j] != '*')
  {
    if (curses)
    {
      move(i++, 0);
      addstr(helptext[j++]);
      if (i >= LINES-1)
      {
        move(i, 0);
        attron(A_STANDOUT);
        addstr("--More--");
        attroff(A_STANDOUT);
        refresh();
        i=0;
        getch();
        clear();
      }
    }
    else
      printf("%s\n", helptext[j++]);
  }
  if (curses)
  {
    refresh();
    endwin();
  }
}

void list_jobs(void)        /* this function is invoked by the -l option */
{                           /* it lists the contents of the queue        */
  static FILE *file;
  char hosts[MAXLENLINE], format[MAXLENLINE], string[MAXLENLINE];
  char line[MAXLENLINE], since[18], when[18];
  char *who, *path, *what, *now, *term;
  int pid, ls, empty=TRUE;
  int cph, cpj, curses=TRUE;
  time_t nowt, s, at;
  register int i=1, j, k;

  make_hostlist(TRUE);

  term=getenv("TERM");                        /* newterm fills LINES and COLS */
  if ((newterm(term, stdout, stdin)) == (SCREEN *)0) curses=FALSE;

  if (curses)
  {
    signal(SIGINT,  terminate);               /* handle quit signals in own   */
    signal(SIGQUIT, terminate);               /* routine                      */
    signal(SIGTERM, terminate);
    cbreak();
    noecho();
    if (COLS < 66)            /* COLS was filled by newterm() with the number */
    {                         /* of columns you have on your screen (window)  */
      addstr("Use a window with more than 65 columns.");
      refresh();
      endwin();
      return;
    }
    else                      /* use this information to adapt the width of   */
    {                         /* the table, cph is columns used for hostnames */
      cph=(COLS-51)*10/29;    /* and cpj is columns used for job              */
      cpj=COLS-51-cph;
    }
  }
  else
  {
    COLS=80;
    cph=10;
    cpj=19;
  }

  while (TRUE)
  {
    if ((file=open_file(NJCPATH, "queue.running", "r")) == NULL) exit(-1);
                                         /* first look into queue.running   */
    nowt=time((time_t *)NULL);
    now=ctime(&nowt);
    now[24]='\0';
    for (j=0; j<COLS-27; j++)
    {
      if (curses)
      {
        move(0,j);
        addstr("-");
      }
      else
        printf("-");
    }
    if (curses)
    {
      move(0, COLS-26);
      addstr(now);
      move(0, COLS-1);
      addstr("-");
    }
    else
      printf(" %s -\n", now);

    if (curses)
    {
      move(1,0);
      addstr("   User      Hosts");
      move(1,cph+13);
      addstr(" S Desired      PID Running/waiting  Job");
    }
    else
      printf("   User      Hosts      S Desired      PID Running/waiting  Job\n");

    if (curses)
    {
      move(2,cph+16);
      addstr("start time            since");
    }
    else
      printf("                          start time            since\n");

    for (j=0; j<COLS; j++)
    {
      if (curses)
      {
        move(3,j);
        addstr("-");
      }
      else
        printf("-");
    }
    if (!curses) printf("\n");

    k=4;
    if (curses) move(k++,0);

    while( (fgets(line,MAXLENLINE,file)) != NULL)     /* and scan through */
    {                                                 /* queue.running    */
      word=strrchr(line,'\n');                        /* extract informations */
      if (word != NULL) *word='\0';
      who=strtok(line, ";");
      strcpy(hosts, strtok(NULL,";"));
      if (*hosts == '*') strcpy(hosts, "ALL");
      if (strlen(hosts) > cph)
        strcpy(hosts+cph-1,">");
      pid=atoi(strtok(NULL,";"));
      s=atoi(strtok(NULL,";"));
      strcpy(when, timetostr(s));
      path=strtok(NULL,";");
      what=strtok(NULL,";");
      if (host_in_list(hosts))
      {
        ls=strcmp(who, user);
        if (strlen(what) > cpj)
          strcpy(what+cpj-1,">");
                                /* and write it to screen if -la was specified, */
        if (all || !ls || !uid) /* it is a job of this user or the user is root */
        {
          empty=FALSE;
          sprintf(format, "%%2d %%-9s %%-%ds R %%16d %%16s%%-%ds\n", cph, cpj);
          sprintf(string, format, i, who, hosts, pid, when, what);
          if (curses)
          {
            addstr(string);
            move(k++,0);
            if (k >= LINES)
            {
              attron(A_STANDOUT);
              addstr("--More--");
              attroff(A_STANDOUT);
              refresh();
              getch();
              clear();
              k=0;
              move(k++,0);
            }
          }
          else
            printf("%s", string);
        }
      }
      i++;
    }
    fclose(file);

    if ((file=open_file(NJCPATH, "queue.waiting", "r")) == NULL) exit(-1);
                                         /* second look into queue.waiting    */
    while( fgets(line, MAXLENLINE, file) != NULL)
    {                                                 /* scan through         */
      word=strrchr(line,'\n');                        /* queue.waiting and    */
      if (word != NULL) *word='\0';                   /* extract informations */
      who=strtok(line, ";");
      strcpy(hosts, strtok(NULL, ";"));
      if (*hosts == '*') strcpy(hosts, "ALL");
      if (strlen(hosts) > cph) strcpy(hosts+cph-1,">");
      at=atoi(strtok(NULL,";"));
      s=atoi(strtok(NULL,";"));
      path=strtok(NULL,";");
      what=strtok(NULL,";");
      if (host_in_list(hosts))
      {
        ls=strcmp(who, user);
        if (strlen(what) > cpj)
          strcpy(what+cpj-1,">");
                                /* and write it to screen if -la was specified, */
        if (all || !ls || !uid) /* it is a job of this user or the user is root */
        {
          empty=FALSE;
          strcpy(since, timetostr(s));
          if (at > 0)
          {
            strcpy(when, timetostr(at));
            if (stopped)
              sprintf(format, "%%2d %%-9s %%-%ds S %%16s%%16s%%-%ds\n", cph, cpj);
            else
              sprintf(format, "%%2d %%-9s %%-%ds W %%16s%%16s%%-%ds\n", cph, cpj);
            sprintf(string, format, i, who, hosts, when, since, what);
            if (curses)
            {
              addstr(string);
              move(k++,0);
              if (k >= LINES)
              {
                attron(A_STANDOUT);
                addstr("--More--");
                attroff(A_STANDOUT);
                refresh();
                getch();
                k=0;
                clear();
                move(k++,0);
              }
            }
            else
              printf("%s", string);
          }
          else
          {
            if (stopped)
              sprintf(format, "%%2d %%-9s %%-%ds S %%16s %%16s %%-%ds\n", cph, cpj);
            else
              sprintf(format, "%%2d %%-9s %%-%ds H %%16s %%16s %%-%ds\n", cph, cpj);
            sprintf(string, format, i, who, hosts, "---", since, what);
            if (curses)
            {
              addstr(string);
              move(k++,0);
            }
            else
              printf("%s", string);
          }
        }
      }
      i++;
    }
    fclose(file);
    if (empty) 
    {
      if (curses)
      {
        move(4, (COLS-13)/2);
        addstr("*** idle ***");
      }
      else
        printf("                                    *** idle ***");
    }
    if (curses) refresh();
    if (!moni) break;
    i=1;
    k=4;
    sleep(10);
  }                                           /* end of while(TRUE) ... */
  if (curses) endwin();
}

void show_stat(void)
{
  char line[MAXLENLINE], format[MAXLENLINE], *term;
  int istat, res, curses=TRUE;
  register int i, j;

  sprintf(format, " %%-%ds %%3d    %%3d    %%3d ", MAXLENHOSTNAME);

  make_hostlist(TRUE);

  term=getenv("TERM");
  if ((newterm(term, stdout, stdin)) == (SCREEN *)0) curses=FALSE;

  if (curses)
  {
    signal(SIGINT,  terminate);         /* handle quit signals in own routine */
    signal(SIGQUIT, terminate);
    signal(SIGTERM, terminate);
    cbreak();
    noecho();
    if (COLS <= MAXLENHOSTNAME+44)
    {
      endwin();
      fprintf(stderr, "%s: ERROR: Use a window with more than %d columns.\n", prgname, MAXLENHOSTNAME+44);
      exit(-1);
    }

    for (i=0; i<=MAXLENHOSTNAME+44; i++)
    {
      move(0, i);
      addstr("-");
    }
    move(1, 0);
    addstr(" Host");
    move(1, MAXLENHOSTNAME+2);
    addstr("Jobs  Maxjobs Nice              Status Load");
    for (i=0; i<=MAXLENHOSTNAME+44; i++)
    {
      move(2, i);
      addstr("-");
    }

    j=3;
    for (i=0; i<=numhosts; i++)
    {
      if (host_in_list(host[i].name))
      {
        sprintf(line, format, host[i].name, host[i].maxjobs, host[i].maxallowed,
                host[i].nice);
        move(j++,0);
        addstr(line);
      }
    }

    if (stopped)
    {
      move(2, (MAXLENHOSTNAME+20)/2);
      attron(A_STANDOUT);
      addstr(" Queue has been stopped ");
      attroff(A_STANDOUT);
    }
    refresh();
  }
  else
  {
    for (i=0; i<=MAXLENHOSTNAME+44; i++) printf("-");
    printf("\n Host");
    for (i=0; i<=MAXLENHOSTNAME-4; i++) printf(" ");
    printf("Jobs  Maxjobs Nice              Status Load\n");
    if (stopped)
    {
      for (i=0; i<=(MAXLENHOSTNAME+20)/2; i++) printf("-");
      printf(" Queue has been stopped ");
      for (i=0; i<(MAXLENHOSTNAME+20)/2; i++) printf("-");
    }
    else
      for (i=0; i<=MAXLENHOSTNAME+44; i++) printf("-");
    printf("\n");
  }

  j=3;
  for (i=0; i<=numhosts; i++)
  {
    if (host_in_list(host[i].name))
    {
      if (curses)
        move(j, MAXLENHOSTNAME+20);
      else
      {
        sprintf(line, format, host[i].name, host[i].maxjobs, host[i].maxallowed,
                host[i].nice);
      }
	if (strcmp(host[i].name, queue_server))
	{
        istat=callrpc(host[i].name, NJCCLIENT, VERS, QUEUECLIENT, (xdrproc_t)xdr_void,
                      (char *)0, (xdrproc_t)xdr_int, (char *)(&res));
      }
	else
      {
        istat=callrpc(host[i].name, NJCSERVER, VERS, QUEUESERVER, (xdrproc_t)xdr_void,
                      (char *)0, (xdrproc_t)xdr_int, (char *)(&res));
	}
      if (istat == RPC_TIMEDOUT)
        if (curses) addstr("Timed out");
        else sprintf(line+MAXLENHOSTNAME+20, "Timed out");
      else if (istat != 0)
        if (curses) addstr("RPC Error");
        else sprintf(line+MAXLENHOSTNAME+20, "RPC Error");
      else if (res < 0)
      {
        if (curses)
        {
          sprintf(line, "Server(%d)", -res);
          addstr(line);
          move(j, MAXLENHOSTNAME+36);
          sprintf(line, "OK  %3d %%", host[i].load);
          addstr(line);
        }
        else 
        {
          sprintf(line+MAXLENHOSTNAME+20, "Server(%d)     ", -res);
          sprintf(line+MAXLENHOSTNAME+36, "OK  %3d %%", host[i].load);
          printf("%s\n", line);
        }
      }
      else
      {
        if (curses)
        {
          sprintf(line, "Client(%d)", res);
          addstr(line);
          move(j, MAXLENHOSTNAME+36);
          sprintf(line, "OK  %3d %%", host[i].load);
          addstr(line);
        }
        else
        {
          sprintf(line+MAXLENHOSTNAME+20, "Client(%d)     ", res);
          sprintf(line+MAXLENHOSTNAME+36, "OK  %3d %%", host[i].load);
          printf("%s\n", line);
        }
      }
      if (curses)
      {
        j++;
        refresh();
      }
    }
  }
  if (curses) endwin();
}

void rm_job(void)
{
  FILE *queue, *j;
  register int i=1;
  char line[MAXLENLINE];

  if (jobno == 0)
  {
    fprintf(stderr, "%s: ERROR: Missing parameter for option -r.\n", prgname);
    exit(-1);
  }

  while (exist(NJCPATH, "lock"))    /* Open queue.running, make sure that */
    sleep(3);                       /* only one process does it           */
  if ((j=open_file(NJCPATH, "lock", "w")) != NULL) fclose(j);

  if ((queue=open_file(NJCPATH, "queue.running", "r")) == NULL)
  {
    rm(NJCPATH, "lock");
    exit(-1);
  }
  while ((fgets(line, MAXLENLINE, queue)) != NULL)
  {
    if (jobno == i++)
    {
      fclose(queue);
      rm(NJCPATH, "lock");
      kill_job(line, TRUE);
      return;
    }
  }
  fclose(queue);

  if ((queue=open_file(NJCPATH, "queue.waiting", "r")) == NULL)
  {
    rm(NJCPATH, "lock");
    exit(-1);
  }
  while ((fgets(line, MAXLENLINE, queue)) != NULL)
  {
    if (jobno == i++)
    {
      fclose(queue);
      rm(NJCPATH, "lock");
      kill_job(line, FALSE);
      return;
    }
  }
  fclose(queue);
  rm(NJCPATH, "lock");
  fprintf(stderr, "%s: ERROR: Unknown job number: %d, check with %s -l.\n",
                   prgname, jobno, prgname);
  exit(-1);
}

void hold_job(char *arg)
{
  FILE *queue, *newqueue;
  char line[MAXLENLINE], l[MAXLENLINE], tmp[L_tmpnam], ts[10], *w;
  int jobs[MAXPROCESS*MAXHOSTS+1], t, all=FALSE;
  register int i=1, j, k;

  if (*arg == '\0')
  {
    fprintf(stderr, "%s: ERROR: Missing parameter for -hold or -rel option.\n", prgname);
    exit(-1);
  }

  while (exist(NJCPATH, "lock"))    /* Open queue.running, make sure that */
    sleep(3);                       /* only one process does it           */
  if ((queue=open_file(NJCPATH, "lock", "w")) != NULL) fclose(queue);

  if ((queue=open_file(NJCPATH, "queue.running", "r")) == NULL)
  {
    rm(NJCPATH, "lock");
    exit(-1);
  }
  while ((fgets(line, MAXLENLINE, queue)) != NULL) i++;
  fclose(queue);
  i--;

  for (j=0; j<=MAXPROCESS*MAXHOSTS+1; j++) jobs[j]=0;

  if (!strcmp(arg, "all"))
    all=TRUE;
  else
  {
    j=0;
    w=word=arg;
    while (*w != '\0')
    {
      if (*w == ',')
      {
        jobs[j++]=atoi(word);
        word=w+1;
      }
      else if (*w == '-')
      {
        jobs[j]=atoi(word);
        w++;
        word=w;
        while (*w != ',' && *w != '\0') w++;
        t=atoi(word);
        while (jobs[j] != t) 
          jobs[j+1]=jobs[j++]+1;
        if (*w != '\0') word=w+1;
      }
      if (*w != '\0') w++;
    }
    jobs[j]=atoi(word);
  }

  j=0;
  while (jobs[j] != 0)
  {
    if (jobs[j++]  < i)
    {
      fprintf(stderr, "%s: ERROR: Unable to use the -hold or -rel option with a running job.\n", prgname);
      exit(-1);
    }
  }

  if ((queue=open_file(NJCPATH, "queue.waiting", "r+")) == NULL)
  {
    rm(NJCPATH, "lock");
    exit(-1);
  }
  if ((newqueue=open_file("", tmp, "w")) == NULL)
  {
    rm(NJCPATH, "lock");
    exit(-1);
  }
  clearerr(newqueue);
  while ((fgets(line, MAXLENLINE, queue)) != NULL)
  {
    *l='\0';
    i++;
    j=0;
    while (jobs[j] != 0 || all)
    {
      if (jobs[j] == i || all)
      {
        word=strtok(line, ";");
        strcpy(l, word);
        strcat(l, ";");
        word=strtok(NULL, ";");
        strcat(l, word);
        strcat(l, ";");
        t=(-atoi(strtok(NULL, ";")));
        sprintf(ts, "%d", t);
        strcat(l, ts);
        strcat(l, ";");
        for (k=1; k<=2; k++)
        {
          word=strtok(NULL, ";");
          strcat(l, word);
          strcat(l, ";");
        }
        word=strtok(NULL, ";");
        strcat(l, word);
        strcpy(line, l);
        jobs[j]=(-1);
        break;
      }   /* end of "if (jobs[j] ... */
      j++;
    }     /* end of "while (jobs[j] ..." */
    fprintf(newqueue, "%s", line);
  }
  j=ferror(newqueue);
  fclose(newqueue);
  if (j)
    error_msg(12, tmp);
  else
    if (!update_file(queue, tmp)) clean_up(0);
  rm(NJCPATH, "lock");

  j=0;
  while (jobs[j] != 0)
  {
    if (jobs[j] != -1)
    {
      fprintf(stderr, "%s: ERROR: Unknown job number: %d, check with %s -l.\n",
                           prgname, jobs[j], prgname);
    }
    j++;
  }
}

void set_queue(int what)
{
  register int i;

  if (what) allowed();

  make_hostlist(TRUE);

  for (i=0; i<=numhosts; i++)
  {
    if (*hostlist[i] != '\0')
      sethost(hostlist[i], what); 
  }
  write_cfg();
}

void submit_job(void)
{
  FILE *file;
  char path[MAXLENLINE];
  char *cwd;
  time_t at, since;
  register int i, j=(-1);

  make_hostlist(TRUE);

  *compi='\0';
  for (i=0; i<=numhosts; i++)
  {
    if (*hostlist[i] != '\0')
    {
      strcat(compi, hostlist[i]);
      strcat(compi, ",");
      j++;
    }
  }
  if (j == numhosts) strcpy(compi, "*");
  if ((word=strrchr(compi, ',')) != NULL) *word='\0';

  if (*attime != '\0')
    at=conv_time();
  else
    at=time((time_t *)NULL);

  since=time((time_t *)NULL);

  if ((cwd=getenv("PWD")) == NULL)    /* use environment variable PWD or cwd to */
  {                                   /* determine current working directory    */
    cwd=getenv("cwd");                /* since getcwd() may return only local   */
  }                                   /* absolute file name                     */

  if (cwd == NULL)
  {
    if (getcwd(path, MAXLENLINE) == NULL)
    {
      fprintf(stderr, "%s: ERROR: Unable to determine current working directory.\n",
              prgname);
      perror((char *)NULL);
      exit(1);
    }
  }
  else
    strcpy(path, cwd);

  while (exist(NJCPATH, "lock"))    /* Open queue.waiting, make sure that */
    sleep(3);                       /* only one process does it           */
  if ((file=open_file(NJCPATH, "lock", "w")) != NULL) fclose(file);

  if ((file=open_file(NJCPATH, "queue.waiting", "a")) == NULL)
  {
    rm(NJCPATH, "lock");
    exit(-1);
  }
  fprintf(file,"%s;%s;%ld;%ld;%s;%s\n",user,compi,at,since,path,job);
  fclose(file);
  rm(NJCPATH, "lock");

  if (stopped)
    fprintf(stderr, "Job submitted, but queue has been stopped.\n");
  else
    force_job();
}

time_t conv_time(void)
{
  struct tm *now;
  time_t t, howmuch=0;
  int dat[5], day=0;

  t=time((time_t *)NULL);
  now=localtime(&t);
  now->tm_isdst=(-1);
  
  word=attime;
  while(*word != '\0')
  {
    *word=tolower(*word);
    word++;
  }

  if (*attime == '+')
  {
    word=attime+1;
    howmuch=atoi(strtok(word, " "));
    word=strtok(NULL, " ");
    if      (!strncmp(word, "mi", 2)) return(t+howmuch*60);
    else if (!strncmp(word, "ho", 2)) return(t+howmuch*3600);
    else if (!strncmp(word, "da", 2)) return(t+howmuch*86400);
    else if (!strncmp(word, "we", 2)) return(t+howmuch*604800);
    else if (!strncmp(word, "mo", 2))
    {
      now->tm_mon += howmuch;
      if (now->tm_mon > 11)
      {
        now->tm_year++;
        now->tm_mon -= 12;
      }
      return(mktime(now));
    }
    else if (!strncmp(word, "ye", 2))
    {
      now->tm_year += howmuch;
      return(mktime(now));
    }
    else
    {
      fprintf(stderr,"%s: ERROR: Unknown time unit: %s.\n", prgname, word);
      exit(-1);
    }
  }
  else if (isalpha(*attime))
  {
    day=now->tm_wday;
    if      (!strncmp(attime,"mon",3))      day=1-day;
    else if (!strncmp(attime,"tue",3))      day=2-day;
    else if (!strncmp(attime,"wed",3))      day=3-day;
    else if (!strncmp(attime,"thu",3))      day=4-day;
    else if (!strncmp(attime,"fri",3))      day=5-day;
    else if (!strncmp(attime,"sat",3))      day=6-day;
    else if (!strncmp(attime,"sun",3))      day=(-day);
    else if (!strncmp(attime,"tomorrow",8)) day=1;
    else if (!strncmp(attime,"hold",4))     return(-t);
    else
    {
      fprintf(stderr,"%s: ERROR: Unknown day of week: %s.\n",prgname,attime);
      exit(-1);
    }
    if (day <= 0) day += 7;

    if (strchr(attime, ',') == NULL)
      return(t+day*86400);
    else
    { 
      t+=day*86400;
      now=localtime(&t);
      strtok(attime, ",");
      now->tm_hour=atoi(strtok(NULL, ":"));
      now->tm_min=atoi(strtok(NULL, ":"));
	now->tm_isdst=(-1);
      return(mktime(now));
    }
  }
  else
  { 
    check_time(attime, dat);
    if (dat[0] != (-1)) now->tm_mon=dat[0]-1;
    if (dat[1] != (-1)) now->tm_mday=dat[1];
    if (dat[2] != (-1)) now->tm_year=dat[2];
    if (dat[3] != (-1)) now->tm_hour=dat[3];
    if (dat[4] != (-1)) now->tm_min=dat[4];
    return(mktime(now));
  }
}

void make_hostlist(int check)
{
  FILE *file;
  register int i, j=(-1);
  char line[MAXLENLINE];

  for (i=0; i<=numhosts; i++)
    *hostlist[i]='\0';

  if (*compi != '*')
  {
    if (*compi == '^')
    {
      for (i=0; i<=numhosts; i++)
        strcpy(hostlist[i], host[i].name);
      word=strtok(compi+1, ",");
      if (check) check_name(word);
      for (i=0; i<=numhosts; i++)
        if (!strcmp(hostlist[i], word)) *hostlist[i]='\0';
      while ((word=strtok(NULL, ",")) != NULL)
      {
        if (check) check_name(word);
        for (i=0; i<=numhosts; i++)
          if (!strcmp(hostlist[i], word)) *hostlist[i]='\0';
      }
    }
    else
    {
      i=0;
      word=strtok(compi,",");
      if (check) check_name(word);
      strcpy(hostlist[i++], word);
      while( (word=strtok(NULL,",")) != NULL)
      {
        if (check) check_name(word);
        strcpy(hostlist[i++], word);
      }
    }
  }
  else
  {
    for (i=0; i<=numhosts; i++)
      strcpy(hostlist[i], host[i].name);
  }

  if ((file=open_file(NJCPATH, "submit.deny", "r")) == NULL) exit(-1);
  while ((fgets(line, MAXLENLINE, file)) != NULL)
  {
    word=strrchr(line, '\n');
    if (word != NULL) *word='\0';
    word=strtok(line, ":");
    if (!strcmp(word, user))
    {
      while ((word=strtok(NULL, " ")) != NULL)
      {
        for (i=0; i<=numhosts; i++)
        {
          if (!strcmp(word, hostlist[i]))
          {
            strcpy(hostlist[i], "\0");
            break;
          }
        }
      }
    }
  }
  fclose(file);

  for (i=0; i<=numhosts; i++)
    if (*hostlist[i] != '\0') j++;

  if (j == -1)
  {
    fprintf(stderr, "%s: ERROR: Permission denied.\n", prgname);
    exit(-1);
  }
}

void check_name(char *word)
{
  char *j;
  int found=FALSE;
  register int i;

  for (j=word; *j != '\0'; j++) *j=tolower(*j);
  for (i=0; i<=numhosts; i++)
  {
    if (!strcmp(word,host[i].name))
    {
      found=TRUE;
      if (host[i].maxjobs == 0)
        fprintf(stderr, "%s: WARNING: Host %s currently does not accept requests.\n", prgname, word);
      break;
    }
  }
  if (!found)
  {
    fprintf(stderr,"%s: ERROR: Unknown host %s.\n", prgname, word);
    exit(-1);
  }
}

char *timetostr(time_t t)
{
  struct tm *s;
  static char str[18];
  static char *day[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };

  s=localtime(&t);

  sprintf(str,"%s %02d/%02d %02d:%02d  ",day[s->tm_wday],s->tm_mon+1,
               s->tm_mday,s->tm_hour,s->tm_min);
  return(str);
}

void sethost(char *which, int what)
{
  register int i;
  int found;

  found=FALSE;
  for (i=0; i<=numhosts; i++)
  {
    if (!strcmp(host[i].name, which))
    { 
      found=TRUE;
      break;
    }
  }
  
  if (!found)
  {
    fprintf(stderr,"%s: ERROR: Unknown host: %s.\n", prgname, which);
    exit(-1);
  }

  if (what == 0)
  {
    if (!uid || (jobs > 0 && jobs <= host[i].maxallowed))
      if (uid && host[i].maxjobs == 0)
        fprintf(stderr, "%s: ERROR: Permission denied for reactivating host %s.\n", prgname, host[i].name);
      else
        host[i].maxjobs=jobs;
    else
    {
      fprintf(stderr, "%s: ERROR: Wrong parameter <n> for option -jobs.\n", prgname);
      fprintf(stderr, "Use 0 < n <= %d\n", host[i].maxallowed);
    }
  }
  else if (what == 1)
    host[i].nice=nicev;
  else if (what == 2)
    host[i].maxallowed=maxjobs;
}

void write_cfg(void)
{
  FILE *cfgfile, *newcfgfile;
  char line[MAXLENLINE], tmp[L_tmpnam];
  register int i;

  if ((cfgfile=open_file(NJCPATH, "queue.cfg", "r+")) == NULL) exit(-1);
  if ((newcfgfile=open_file("", tmp, "w")) == NULL) exit(-1);
  clearerr(newcfgfile);
  while ((fgets(line, MAXLENLINE, cfgfile)) != NULL)
  {
    if (*line == '#')
      fprintf(newcfgfile, "%s", line);
    else
    {
      word=strtok(line, " \t");
      for (i=0; i<=numhosts; i++)
      {
        if (!strcmp(host[i].name, word))
        {
          strtok(NULL, " \t");
          strtok(NULL, " \t");
          fprintf(newcfgfile, "%s\t%d\t%d\t%d\n", host[i].name,
                  host[i].maxjobs, host[i].maxallowed, host[i].nice);
          break;
        }
      }
    }
  }
  i=ferror(newcfgfile);
  fclose(newcfgfile);
  if (i)
  {
    error_msg(12, tmp);
    clean_up(0);
  }
  if (!update_file(cfgfile, tmp)) clean_up(0);

  if ((i=callrpc(queue_server, NJCSERVER, VERS, INIT, (xdrproc_t)xdr_void, (char *)0,
                 (xdrproc_t)xdr_void, (char *)0)) != 0)
  {
    fprintf(stderr, "%s: ERROR: Cannot contact queue server %s.\n", prgname, queue_server);
    clnt_perrno(i);
    fprintf(stderr, "\n");
  }
}

void allowed(void)
{
  if (uid)
  {
    fprintf(stderr, "%s: ERROR: Permission denied.\n", prgname);
    exit(-1);
  }
} 

void kill_job(char *str, int isrun)
{
  FILE *queue, *newqueue;
  char l1[MAXLENLINE], l2[MAXLENLINE], tmp[MAXLENLINE], *where, *pid;
  register int i;

  strcpy(l2, str);
  word=strrchr(l2, '\n');
  *word='\0';
  word=strtok(str, ";");
  if (uid)
  {
    if (strcmp(user, word))
    {
      fprintf(stderr, "%s: ERROR: Permission denied.\n", prgname);
      exit(-1);
    }
  }
  if (isrun)
  {
    where=strtok(NULL, ";");
    pid=strtok(NULL, ";");

    strcpy(l1, "rsh ");
    strcat(l1, where);
    strcat(l1, " ");
    strcat(l1, NJCPATH);
    strcat(l1, "/rmjob ");
    strcat(l1, pid);

    system(l1);
  }
  else
  {
    while (exist(NJCPATH, "lock"))    /* Open queue.waiting, make sure that */
      sleep(3);                       /* only one process does it           */
    if ((queue=open_file(NJCPATH, "lock", "w")) != NULL) fclose(queue);

    if ((queue=open_file(NJCPATH, "queue.waiting", "r+")) == NULL)
    {
      rm(NJCPATH, "lock");
      exit(-1);
    }
    if ((newqueue=open_file("", tmp, "w")) != NULL)   /* Open temporary file */
    {
      clearerr(newqueue);
      while ((fgets(l1, MAXLENLINE, queue)) != NULL)
      {
        if (strncmp(l1, l2, strlen(l2)))   /* Copy all lines from queue.waiting  */
          fprintf(newqueue, "%s", l1);     /* to the temporary file except the   */
      }                                    /* line with the job which is to be   */
      i=ferror(newqueue);                  /* deleted                            */ 
      fclose(newqueue);
      if (i)
      {
        error_msg(12, tmp);
        clean_up(0);
      }
      if (!update_file(queue, tmp)) clean_up(0); /* copy back from temporary file */
    }
    rm(NJCPATH, "lock");
  }
}

void stop_queue(void)
{
  int istat;

  allowed();

  if ((istat=callrpc(queue_server, NJCSERVER, VERS, STOP, (xdrproc_t)xdr_void,
                     (char *)0, (xdrproc_t)xdr_void, (char *)0)) != 0)
  {
    fprintf(stderr, "%s: ERROR: Cannot contact queue server %s.\n", prgname,
            queue_server);
    clnt_perrno(istat);
    fprintf(stderr, "\n");
  }
}

void force_job(void)
{
  int istat;

  if ((istat=callrpc(queue_server, NJCSERVER, VERS, NEXTJOB, (xdrproc_t)xdr_void,
                     (char *)0, (xdrproc_t)xdr_void, (char *)0)) != 0)
  {
    fprintf(stderr, "%s: ERROR: Cannot contact queue server %s.\n", prgname,
            queue_server);
    clnt_perrno(istat);
    fprintf(stderr, "\n");
  }
}

void on_off(int what)
{
  FILE *cfgfile, *newcfgfile;
  char line[MAXLENLINE], saveline[MAXLENLINE], tmp[L_tmpnam], *l;
  register int i, found;

  allowed();

  make_hostlist(FALSE);

  if ((cfgfile=open_file(NJCPATH, "queue.cfg", "r+")) == NULL) exit(-1);
  if ((newcfgfile=open_file("", tmp, "w")) == NULL) exit(-1);
  clearerr(newcfgfile);
  while ((fgets(line, MAXLENLINE, cfgfile)) != NULL)
  {
    strcpy(saveline, line);
    l=saveline;
    if (what)
    {
      if (*line == '#' && *(line+1) == '#')
      {
        l+=2;
        word=strtok(line+2," \t");
        found=FALSE;
        for (i=0; i<=numhosts; i++)
        {
          if (!strcmp(hostlist[i], word))
          {
            found=TRUE;
            break;
          }
        }
        if (!found) l-=2;
      }
      fprintf(newcfgfile, "%s", l);
    }
    else
    {
      if (*line == '#')
        fprintf(newcfgfile, "%s", l);
      else
      {
        word=strtok(line," \t");
        found=FALSE;
        for (i=0; i<=numhosts; i++)
        {
          if (!strcmp(hostlist[i], word))
          {
            fprintf(newcfgfile, "##%s", l);
            found=TRUE;
            break;
          }
        }
        if (!found) fprintf(newcfgfile, "%s", l);
      }
    }    /* end of if (what) ... */
  }      /* end of while loop */
  i=ferror(newcfgfile);
  fclose(newcfgfile);
  if (i)
  {
    error_msg(12, tmp);
    clean_up(0);
  }
  if (!update_file(cfgfile, tmp)) clean_up(0);

  if ((i=callrpc(queue_server, NJCSERVER, VERS, INIT, (xdrproc_t)xdr_void, (char *)0,
                 (xdrproc_t)xdr_void, (char *)0)) != 0)
  {
    fprintf(stderr, "%s: ERROR: Cannot contact queue server %s.\n", prgname, queue_server);
    clnt_perrno(i);
    fprintf(stderr, "\n");
  }
} 

void check_time(char *str, int dat[5])
{
  char *word;
  register int i;

  for (i=0; i<5; i++) dat[i]=(-1);
  dat[0]=atoi(strtok(str, "/,:"));
  i=1;
  while ((word=strtok(NULL, "/,:")) != NULL) dat[i++]=atoi(word);

  switch (i)
  {
    case 2: dat[3]=dat[0];
		dat[4]=dat[1];
		dat[0]=dat[1]=(-1);
		break;
    case 3: fprintf(stderr, "%s: ERROR: Day is missing in parameter of -t option.\n", prgname);
		exit(-1);
		break;
    case 4: dat[4]=dat[3];
            dat[3]=dat[2];
            dat[2]=(-1);
		break;
  }

  if ((dat[0] < 1 || dat[0] > 12) && dat[0] != (-1))
  {
    fprintf(stderr, "%s: ERROR: Month is out of range in parameter of -t option: %d.\n", prgname, dat[0]);
    exit(-1);
  }
  if ((dat[1] < 1 || dat[1] > 31) && dat[1] != (-1))
  {
    fprintf(stderr, "%s: ERROR: Day is out of range in parameter of -t option: %d.\n", prgname, dat[1]);
    exit(-1);
  }
  if ((dat[3] < 0 || dat[3] > 23) && dat[3] != (-1))
  {
    fprintf(stderr, "%s: ERROR: Hour is out of range in parameter of -t option: %d.\n", prgname, dat[3]);
    exit(-1);
  }
  if ((dat[4] < 0 || dat[4] > 59) && dat[4] != (-1))
  {
    fprintf(stderr, "%s: ERROR: Minute is out of range in parameter of -t option: %d.\n", prgname, dat[4]);
    exit(-1);
  }
}

void interactive(void)
{
  int choosen;
  char line[MAXLENLINE], *term;
  static char *menue[] = { "Submit a job",
                           "Force queue to try to start the next job",
                           "Remove a job from the queue",
                           "List your jobs in the queue",
                           "List all jobs in the queue",
                           "Show parameters of the queue",
                           "Set a job to hold or release a holded job",
                           "Set number of jobs for queue",
                           "Get help",
                           "Set maximum number of jobs for queue",
                           "Set nice value for queue",
                           "Start or stop the queue"
                         };

  term=getenv("TERM");
  if ((newterm(term, stdout, stdin)) == (SCREEN *)0)
  {
    helpme();
    exit(0);
  }
  signal(SIGINT,  terminate);           /* handle quit signals in own routine */
  signal(SIGQUIT, terminate);
  signal(SIGTERM, terminate);
  init();
  cbreak();
  noecho();
 
  if (!uid)
    choosen=menu(menue, 11, "SUBMIT - Choose one of the following options:", 1, 1);
  else
    choosen=menu(menue,  8, "SUBMIT - Choose one of the following options:", 1, 1);

  clear();
  endwin();

  switch (choosen)
  {
   case -1: fprintf(stderr, "%s: ERROR: Use a larger window.\n", prgname);
            break;
    case 1: printf("Job: ");
            fgets(job, MAXLENLINE, stdin);
            word=strrchr(job, '\n');
            if (word != NULL) *word='\0';
            printf("Hostnames (separated by colons, default ALL): ");
            fgets(compi, MAXLENLINE, stdin);
            if (*compi == 10) *compi='*';
            word=strrchr(compi, '\n');
            if (word != NULL) *word='\0';
            printf("Start at (default now): ");
            fgets(attime, MAXLENLINE, stdin);
            if (*attime == 10) *attime='\0';
            word=strrchr(attime, '\n');
            if (word != NULL) *word='\0';
            submit_job();
            break;
    case 2: force_job();
            break;
    case 3: list=TRUE;
            all=TRUE;
            list_jobs();
            printf("\nEnter number of job to be removed: ");
            scanf("%d", &jobno);
            rm_job();
            break;
    case 4: list=TRUE;
            list_jobs();
            break;
    case 5: list=TRUE;
            all=TRUE;
            list_jobs();
            break;
    case 6: status=TRUE;
            show_stat();
            break;
    case 7: list=TRUE;
            all=TRUE;
            list_jobs();
            printf("\nEnter number of job to be holded or released: ");
            scanf("%s", line);
            hold_job(line);
            break;
    case 8: printf("Number of jobs: ");
            scanf("%d", &jobs);
            printf("Hostnames (separated by colons, default ALL): ");
            fgets(compi, MAXLENLINE, stdin);
            if (*compi == 10) *compi='*';
            word=strrchr(compi, '\n');
            if (word != NULL) *word='\0';
            set_queue(0);
            break;
    case 9: helpme();
            break;
   case 10: printf("Number of jobs: ");
            scanf("%d", &jobs);
            printf("Hostnames (separated by colons, default ALL): ");
            fgets(compi, MAXLENLINE, stdin);
            if (*compi == 10) *compi='*';
            word=strrchr(compi, '\n');
            if (word != NULL) *word='\0';
            set_queue(2);
            break;
   case 11: printf("New nice value: ");
            scanf("%d", &nicev);
            printf("Hostnames (separated by colons, default ALL): ");
            fgets(compi, MAXLENLINE, stdin);
            if (*compi == 10) *compi='*';
            word=strrchr(compi, '\n');
            if (word != NULL) *word='\0';
            set_queue(1);
            break;
   case 12: stop_queue();
            break;
  }
  exit(0);
}

void clean_up(int dummy)
{
  if (exist(NJCPATH, "lock")) rm(NJCPATH, "lock");
  ende(-1);
}

void ende(int i)
{
  exit(i);
}

int host_in_list(char *hosts)
{
  char help[MAXLENLINE];
  char *start[MAXHOSTS];
  register int i=0, j, count=0;

  if (!strcmp(hosts, "ALL")) return(TRUE);

  strcpy(help, hosts);
  start[count]=strtok(help, ",");
  while((start[++count]=strtok(NULL, ",")) != NULL);

  i=0;
  while(*hostlist[i] != '\0')
  {
    for (j=0; j<=count-1; j++)
      if (!strcmp(hostlist[i++], start[j])) return(TRUE);
  }
  return(FALSE);
}

void terminate(int dummy)
{
  endwin();
  exit(-1);
}

int menu(char **menue, int n, char *title, int x, int y)
{
  int i, selected=0, key;

  if (LINES < n+y) return(-1);
  for (i=0; i<=n; i++)
  {
    if (COLS < strlen(menue[i])+x) return(-1);
  }

  keypad(stdscr,TRUE);
  nodelay(stdscr, FALSE);

  if (title != NULL)
  {
    move(y, x);
    addstr(title);
    y++;
  }

  while(TRUE)
  {
    for (i=0; i<=n; i++)
    {
      move(y+i, x);
      if (i==selected) attron(A_STANDOUT);
      addstr(menue[i]);
      if (i==selected) attroff(A_STANDOUT);
    }
    move(LINES-1, COLS-1);
    refresh();

    key=getch();
    switch(key)
    {
      case KEY_DOWN: selected++;
                     break;
      case KEY_UP:   selected--;
                     break;
      default:       return(selected+1);
                     break;
    }
    if (selected < 0) selected=n;
    if (selected > n) selected=0;
  }
}
